@@ -12,12 +12,96 @@ import Data.Data
 import Data.Generics
 import Data.Maybe
 import Data.List
-import qualified Data.IntSet as IntSet
-import Data.IntSet(IntSet)
-import qualified Data.IntMap as IntMap
-import Data.IntMap(IntMap)
 import Data.IORef
 import Control.Exception
+import Control.Monad
+import System.Environment(getEnv)
+import qualified Data.IntMap as IntMap; import Data.IntMap(IntMap)
+
+
+#if __GLASGOW_HASKELL__ < 606
+
+---------------------------------------------------------------------
+-- GHC 6.4 and below
+
+import qualified Data.Set as Set
+import qualified Data.Map as Map
+
+type TypeKey = TypeRep
+type TypeSet = Set.Set TypeKey
+type TypeMap = Map.Map TypeKey
+
+typeKey :: Typeable a => a -> TypeKey
+typeKey = typeOf
+
+#elif __GLASGOW_HASKELL__ < 702
+
+---------------------------------------------------------------------
+-- GHC 6.6 to 7.0 (has typeRepKey)
+
+import qualified Data.IntSet as Set
+import qualified Data.IntMap as Map
+
+type TypeKey = Int
+type TypeSet = Set.IntSet
+type TypeMap = Map.IntMap
+
+typeKey :: Typeable a => a -> TypeKey
+typeKey x = inlinePerformIO $ typeRepKey $ typeOf x
+
+#else
+
+---------------------------------------------------------------------
+-- GHC 7.2 and above (using fingerprint)
+
+import qualified Data.HashMap.Strict as Map
+import qualified Data.HashSet as Set
+type TypeSet = Set.HashSet TypeKey
+type TypeMap = Map.HashMap TypeKey
+
+type TypeKey = TypeRep
+
+typeKey :: Typeable a => a -> TypeKey
+typeKey = typeOf
+
+#endif
+
+
+#if __GLASGOW_HASKELL__ < 702
+
+---------------------------------------------------------------------
+-- GHC 7.0 and below (using containers API)
+
+(!) = (Map.!)
+map_findWithDefault = Map.findWithDefault
+map_fromAscList = Map.fromAscList
+map_keysSet = Map.keysSet
+map_member = Map.member
+set_partition = Set.partition
+set_toAscList = Set.toAscList
+set_unions = Set.unions
+
+#else
+
+---------------------------------------------------------------------
+-- GHC 7.2 and above (using unordered-containers API)
+
+(!) mp k = Map.lookupDefault (error "Could not find element") k mp
+map_findWithDefault = Map.lookupDefault
+map_fromAscList = Map.fromList
+map_keysSet = Set.fromList . Map.keys
+map_member x xs = isJust $ Map.lookup x xs
+set_partition f x = (Set.filter f x, Set.filter (not . f) x)
+set_toAscList = Set.toList
+set_unions = foldr Set.union Set.empty
+
+#endif
+
+
+{-# NOINLINE uniplateVerbose #-}
+uniplateVerbose :: Int -- -1 = error if failed, 0 = quiet, 1 = print errors only, 2 = print everything
+uniplateVerbose = unsafePerformIO $ do
+    fmap read (getEnv "UNIPLATE_VERBOSE") `Control.Exception.catch` \(_ :: SomeException) -> return 0
 
 
 ---------------------------------------------------------------------
@@ -32,126 +116,6 @@ data Oracle to = Oracle {fromOracle :: forall on . Typeable on => on -> Answer t
 
 {-# INLINE hitTest #-}
 hitTest :: (Data from, Data to) => from -> to -> Oracle to
-
-
-#if __GLASGOW_HASKELL__ < 606
--- GHC 6.4.2 does not export typeRepKey, so we can't do the trick
--- as efficiently, so we just give up and revert to always following
-
-hitTest _ _ = Oracle . maybe Follow Hit . cast
-
-
-#elif 0
-
-
-hitTest from to =
-    let kto = typeKey to
-    in case hitTestQuery (dataBox from) kto of
-           Nothing -> Oracle $ \on -> if typeKey on == kto then Hit $ unsafeCoerce on else Follow
-           Just cache -> let test = cacheHitTest cache in
-               Oracle $ \on -> let kon = typeKey on in
-                   if kon == kto then Hit $ unsafeCoerce on
-                   else if test kon then Follow
-                   else Miss
-
-
--- A cache hit test, but partially evaluated
-{-# INLINE cacheHitTest #-}
-cacheHitTest :: Cache -> TypeKey -> Bool
-cacheHitTest (Cache hit miss)
-    | IntSet.null hit = const False
-    | IntSet.null miss = const True
-    | otherwise = \x -> x `IntSet.member` hit
-
-
--- hit means that this value may result in a hit
--- miss means that this value will never result in a hit
-data Cache = Cache {hit :: IntSet, miss :: IntSet} deriving Show
-newCache = Cache IntSet.empty IntSet.empty
-
-
--- Indexed by the @from@ type, then the @to@ type
--- Nothing means that we can't perform the trick on the set
-{-# NOINLINE hitTestCache #-}
-hitTestCache :: IORef (IntMap (IntMap (Maybe Cache)))
-hitTestCache = unsafePerformIO $ newIORef IntMap.empty
-
-
-hitTestQuery :: DataBox -> TypeKey -> Maybe Cache
-hitTestQuery from@(DataBox kfrom vfrom) kto = inlinePerformIO $ do
-    mp <- readIORef hitTestCache
-    let res = IntMap.lookup kfrom mp >>= IntMap.lookup kto
-    case res of
-        Just ans -> return ans
-        Nothing -> do
-            let res = toCache $ hitTestAdd from kto IntMap.empty
-            res2 <- Control.Exception.catch (return $! res) (\(_ :: SomeException) -> return Nothing)
-            -- -- uncomment these lines to see where type search fails
-            -- if isNothing res2 then print ("failure",show (typeOf vfrom),kfrom,kto) else return ()
-
-            atomicModifyIORef hitTestCache $ \mp -> flip (,) () $
-                IntMap.insertWith (const $ IntMap.insert kto res2) kfrom (IntMap.singleton kto res2) mp
-            return res2
-
-
--- need to classify each item as one of the following
-data Res = RHit | RMiss | RFollow | RBad deriving (Show,Eq)
-
-
-toCache :: IntMap Res -> Maybe Cache
-toCache res | not $ IntSet.null $ f RBad = Nothing
-            | otherwise = Just $ Cache (f RFollow) (f RMiss)
-    where f x = IntMap.keysSet $ IntMap.filter (== x) res
-
-hitTestAdd :: DataBox -> TypeKey -> IntMap Res -> IntMap Res
-hitTestAdd from@(DataBox kfrom _) kto res = case sybChildren from of
-    _ | kfrom `IntMap.member` res -> res
-    Nothing -> IntMap.insert kfrom RBad res
-
-    -- make an inductive hypothesis that this value is a miss
-    -- if it turns out you were wrong, start again
-    -- uses backtracking, so could be expensive
-    Just xs | kto == kfrom -> hitTestAdds xs kto $ IntMap.insert kfrom RHit res
-            | correct -> res2
-            | otherwise -> hitTestAdds xs kto $ IntMap.insert kfrom RFollow res
-        where res2 = hitTestAdds xs kto $ IntMap.insert kfrom RMiss res
-              correct = all ((==) RMiss . (res2 IntMap.!) . dataBoxKey) xs
-
-hitTestAdds :: [DataBox] -> TypeKey -> IntMap Res -> IntMap Res
-hitTestAdds [] kto res = res
-hitTestAdds (x:xs) kto res = hitTestAdds xs kto $ hitTestAdd x kto res
-
-
-type TypeKey = Int
-
-typeKey :: Typeable a => a -> Int
-typeKey x = inlinePerformIO $ typeRepKey $ typeOf x
-
-
--- | An existential box representing a type which supports SYB
--- operations.
-data DataBox = forall a . (Data a) => DataBox {dataBoxKey :: TypeKey, dataBoxVal :: a}
-
-dataBox :: Data a => a -> DataBox
-dataBox x = DataBox (typeKey x) x
-
--- return all the possible children of a node
--- if you can't do so, just return Nothing
-sybChildren :: DataBox -> Maybe [DataBox]
-sybChildren (DataBox k x)
-    | k == typeRational = Just [dataBox (0 :: Integer)]
-    | isAlgType dtyp = Just $ concatMap f ctrs
-    | isNorepType dtyp = Nothing
-    | otherwise = Just []
-    where
-        f ctr = gmapQ dataBox (asTypeOf (fromConstr ctr) x)
-        ctrs = dataTypeConstrs dtyp
-        dtyp = dataTypeOf x
-
-typeRational = typeKey (undefined :: Rational)
-
-#else
-
 hitTest from to =
     let kto = typeKey to
     in case readCacheFollower (dataBox from) kto of
@@ -167,13 +131,13 @@ hitTest from to =
 -- CACHE
 -- Store and compute the Follower and HitMap
 
-data Cache = Cache HitMap (IntMap2 (Maybe Follower))
+data Cache = Cache HitMap (TypeMap2 (Maybe Follower))
 
 -- Indexed by the @from@ type, then the @to@ type
 -- Nothing means that we can't perform the trick on the set
 {-# NOINLINE cache #-}
 cache :: IORef Cache
-cache = unsafePerformIO $ newIORef $ Cache emptyHitMap IntMap.empty
+cache = unsafePerformIO $ newIORef $ Cache emptyHitMap Map.empty
 
 
 readCacheFollower :: DataBox -> TypeKey -> Maybe Follower
@@ -182,12 +146,17 @@ readCacheFollower from@(DataBox kfrom vfrom) kto = inlinePerformIO $ do
     case lookup2 kfrom kto follow of
         Just ans -> return ans
         Nothing -> do
-            res <- Control.Exception.catch (return $! Just $! insertHitMap from hit) (\(_ :: SomeException) -> return Nothing)
+            res <- Control.Exception.try (return $! insertHitMap from hit)
             (hit,fol) <- return $ case res of
-                Nothing -> (hit, Nothing)
-                Just hit -> (hit, Just $ follower kfrom kto hit)
-            -- -- uncomment these lines to see where type search fails
-            -- if isNothing fol then print ("failure",show (typeOf vfrom),kfrom,kto) else return ()
+                Left _ -> (hit, Nothing)
+                Right hit -> (hit, Just $ follower kfrom kto hit)
+
+            let msg =
+                    "# Uniplate lookup on (" ++ show (typeOf vfrom) ++ "), from (" ++ show kfrom ++ "), to (" ++ show kto ++ "): " ++
+                    either (\(msg::SomeException) -> "FAILURE (" ++ show msg ++ ")") (const "Success") res
+
+            when (uniplateVerbose + maybe 1 (const 0) fol >= 2) $ putStrLn msg
+            when (uniplateVerbose < 0 && isNothing fol) $ error msg
 
             atomicModifyIORef cache $ \(Cache _ follow) -> (Cache hit (insert2 kfrom kto fol follow), ())
             return fol
@@ -197,7 +166,7 @@ readCacheFollower from@(DataBox kfrom vfrom) kto = inlinePerformIO $ do
 readCacheHitMap :: DataBox -> Maybe HitMap
 readCacheHitMap from@(DataBox kfrom vfrom) = inlinePerformIO $ do
     Cache hit _ <- readIORef cache
-    case IntMap.lookup kfrom hit of
+    case Map.lookup kfrom hit of
         Just _ -> return $ Just hit
         Nothing -> do
             res <- Control.Exception.catch (return $! Just $! insertHitMap from hit) (\(_ :: SomeException) -> return Nothing)
@@ -209,15 +178,24 @@ readCacheHitMap from@(DataBox kfrom vfrom) = inlinePerformIO $ do
 
 
 ---------------------------------------------------------------------
--- INTMAP2
+-- TYPEMAP2/INTMAP2
+
+type TypeMap2 a = TypeMap (TypeMap a)
+
+lookup2 :: TypeKey -> TypeKey -> TypeMap2 a -> Maybe a
+lookup2 x y mp = Map.lookup x mp >>= Map.lookup y
+
+insert2 :: TypeKey -> TypeKey -> a -> TypeMap2 a -> TypeMap2 a
+insert2 x y v mp = Map.insertWith (const $ Map.insert y v) x (Map.singleton y v) mp
+
 
 type IntMap2 a = IntMap (IntMap a)
 
-lookup2 :: Int -> Int -> IntMap (IntMap x) -> Maybe x
-lookup2 x y mp = IntMap.lookup x mp >>= IntMap.lookup y
+intLookup2 :: Int -> Int -> IntMap2 a -> Maybe a
+intLookup2 x y mp = IntMap.lookup x mp >>= IntMap.lookup y
 
-insert2 :: Int -> Int -> x -> IntMap (IntMap x) -> IntMap (IntMap x)
-insert2 x y v mp = IntMap.insertWith (const $ IntMap.insert y v) x (IntMap.singleton y v) mp
+intInsert2 :: Int -> Int -> a -> IntMap2 a -> IntMap2 a
+intInsert2 x y v mp = IntMap.insertWith (const $ IntMap.insert y v) x (IntMap.singleton y v) mp
 
 
 ---------------------------------------------------------------------
@@ -230,23 +208,18 @@ type Follower = TypeKey -> Bool
 -- HitMap must have addHitMap on the key
 follower :: TypeKey -> TypeKey -> HitMap -> Follower
 follower from to mp
-    | IntSet.null hit = const False
-    | IntSet.null miss = const True
-    | otherwise = \now -> now `IntSet.member` hit
+    | Set.null hit = const False
+    | Set.null miss = const True
+    | Set.size hit < Set.size miss = \k -> k `Set.member` hit
+    | otherwise = \k -> not $ k `Set.member` miss
     where
-        (hit,miss) = IntSet.partition (\x -> to `IntSet.member` grab x) (IntSet.insert from $ grab from)
-        grab x = IntMap.findWithDefault (error "couldn't grab in follower") x mp
+        (hit,miss) = set_partition (\x -> to `Set.member` grab x) (Set.insert from $ grab from)
+        grab x = map_findWithDefault (error "couldn't grab in follower") x mp
 
 
 ---------------------------------------------------------------------
 -- DATA/TYPEABLE OPERATIONS
 
-type TypeKey = Int
-
-typeKey :: Typeable a => a -> Int
-typeKey x = inlinePerformIO $ typeRepKey $ typeOf x
-
-
 -- | An existential box representing a type which supports SYB
 -- operations.
 data DataBox = forall a . (Data a) => DataBox {dataBoxKey :: TypeKey, dataBoxVal :: a}
@@ -259,7 +232,11 @@ dataBox x = DataBox (typeKey x) x
 sybChildren :: Data a => a -> [DataBox]
 sybChildren x
     | isAlgType dtyp = concatMap f ctrs
-    | isNorepType dtyp = error "sybChildren on NorepType"
+    | isNorepType dtyp = []
+        -- Extensive discussions with Lennart and Roman decided that if something returns NorepType, it really wants to be atomic
+        -- so we should let it be, and pretend it has no children.
+        -- The most common types which say this are Data.Set/Data.Map, and we think that's a bug in their Data instances.
+        -- error $ "Data.Generics.Uniplate.Data: sybChildren on data type which returns NorepType, " ++ show (typeOf x) ++ ", " ++ show dtyp
     | otherwise = []
     where
         f ctr = gmapQ dataBox (asTypeOf (fromConstr ctr) x)
@@ -271,26 +248,26 @@ sybChildren x
 -- HITMAP
 -- What is the transitive closure of a type key
 
-type HitMap = IntMap IntSet
+type HitMap = TypeMap TypeSet
 
 emptyHitMap :: HitMap
-emptyHitMap = IntMap.fromList
-        [(tRational, IntSet.singleton tInteger)
-        ,(tInteger, IntSet.empty)]
+emptyHitMap = Map.fromList
+        [(tRational, Set.singleton tInteger)
+        ,(tInteger, Set.empty)]
     where tRational = typeKey (undefined :: Rational)
           tInteger = typeKey (0 :: Integer)
 
 
 insertHitMap :: DataBox -> HitMap -> HitMap
-insertHitMap box hit = fixEq trans (populate box) `IntMap.union` hit
+insertHitMap box hit = fixEq trans (populate box) `Map.union` hit
     where
         -- create a fresh box with all the necessary children that aren't in hit
         populate :: DataBox -> HitMap
-        populate x = f x IntMap.empty
+        populate x = f x Map.empty
             where
                 f (DataBox key val) mp
-                    | key `IntMap.member` hit || key `IntMap.member` mp = mp
-                    | otherwise = fs cs $ IntMap.insert key (IntSet.fromList $ map dataBoxKey cs) mp
+                    | key `map_member` hit || key `map_member` mp = mp
+                    | otherwise = fs cs $ Map.insert key (Set.fromList $ map dataBoxKey cs) mp
                         where cs = sybChildren val
 
                 fs [] mp = mp
@@ -299,10 +276,10 @@ insertHitMap box hit = fixEq trans (populate box) `IntMap.union` hit
 
         -- update every one to be the transitive closure
         trans :: HitMap -> HitMap
-        trans mp = IntMap.map f mp
+        trans mp = Map.map f mp
             where
-                f x = IntSet.unions $ x : map g (IntSet.toList x)
-                g x = IntMap.findWithDefault (hit IntMap.! x) x mp
+                f x = set_unions $ x : map g (Set.toList x)
+                g x = map_findWithDefault (hit ! x) x mp
 
 
 fixEq :: Eq a => (a -> a) -> a -> a
@@ -310,9 +287,6 @@ fixEq f x = if x == x2 then x2 else fixEq f x2
     where x2 = f x
 
 
-#endif
-
-
 ---------------------------------------------------------------------
 -- INSTANCE FUNCTIONS
 
@@ -382,9 +356,6 @@ transformBis = transformBis_
 
 transformBis_ :: forall a . Data a => [[Transformer]] -> a -> a
 
-
-#if __GLASGOW_HASKELL__ >= 606
-
 -- basic algorithm:
 -- as you go down, given transformBis [fN..f1]
 --   if x is not in the set reachable by fN..f1, return x
@@ -397,39 +368,37 @@ transformBis_ ts | isJust hitBoxM = op (sliceMe 1 n)
         on = dataBox (undefined :: a)
         hitBoxM = readCacheHitMap on
         hitBox = fromJust hitBoxM
-        univ = IntSet.toAscList $ IntSet.insert (dataBoxKey on) $ hitBox IntMap.! dataBoxKey on
+        univ = set_toAscList $ Set.insert (dataBoxKey on) $ hitBox ! dataBoxKey on
         n = length ts
 
         -- (a,b), where a < b, and both in range 1..n
-        sliceMe i j = fromMaybe IntMap.empty $ lookup2 i j slices
-        slices :: IntMap2 (IntMap (Maybe Transformer))
+        sliceMe i j = fromMaybe Map.empty $ intLookup2 i j slices
+        slices :: IntMap2 (TypeMap (Maybe Transformer))
         slices = IntMap.fromAscList
             [ (i, IntMap.fromAscList [(j, slice i j ts) | (j,ts) <- zip [i..n] (tail $ inits ts)])
             | (i,ts) <- zip [1..n] (tails $ reverse ts)]
 
-        slice :: Int -> Int -> [[Transformer]] -> IntMap (Maybe Transformer)
+        slice :: Int -> Int -> [[Transformer]] -> TypeMap (Maybe Transformer)
         slice from to tts = self
             where
-                self = f IntMap.empty (zip [from..] tts) -- FIXME: flattening out here gives different results...
+                self = f Map.empty (zip [from..] tts) -- FIXME: flattening out here gives different results...
                 f a ((i,[Transformer tk tr]):ts)
-                    | tk `IntMap.member` a = f a ts
-                    | otherwise = f (IntMap.insert tk t a) ts
+                    | tk `map_member` a = f a ts
+                    | otherwise = f (Map.insert tk t a) ts
                     where
                         t = Just $ Transformer tk $ op (sliceMe (i+1) to) . tr . gmapT (op $ sliceMe from i)
 
-                f a [] = a `IntMap.union` IntMap.fromAscList (mapMaybe (g $ IntMap.keysSet a) $ univ)
+                f a [] = a `Map.union` map_fromAscList (mapMaybe (g $ map_keysSet a) univ)
 
                 g a t = if b then Nothing else Just (t, Nothing)
-                    where b = IntSet.null $ a `IntSet.intersection` (hitBox IntMap.! t)
+                    where b = Set.null $ a `Set.intersection` (hitBox ! t)
 
-        op :: forall b . Data b => IntMap (Maybe Transformer) -> b -> b
-        op slice = case IntMap.lookup (typeKey (undefined :: b)) slice of
+        op :: forall b . Data b => TypeMap (Maybe Transformer) -> b -> b
+        op slice = case Map.lookup (typeKey (undefined :: b)) slice of
             Nothing -> id
             Just Nothing -> gmapT (op slice)
             Just (Just (Transformer _ t)) -> unsafeCoerce . t . unsafeCoerce
 
-#endif
-
 
 transformBis_ [] = id
 transformBis_ ([]:xs) = transformBis_ xs
